#!/usr/bin/env python
from anki.collection import Collection, ExportAnkiPackageOptions, ImportCsvRequest, ImportAnkiPackageOptions, ImportAnkiPackageRequest, Delimiter, CsvMetadata, DupeResolution
from anki.generic_pb2 import StringList
from anki.import_export_pb2 import ImportAnkiPackageUpdateCondition
from anki.models import ChangeNotetypeRequest
from pathlib import Path
from pydantic import BaseModel, field_serializer, field_validator
from semver import Version
from tempfile import TemporaryDirectory
from typing import Literal
from yaml import load, CLoader
import argparse
import random

SPEC_FILENAME = '.apkg-spec.yaml'
parser = argparse.ArgumentParser(
    prog="export_apkgs.py",
    description="Generates apkgs from a spec in a specified directory")
parser.add_argument('-c', '--content', default='content', help="path to YAML or directory tree that contains .apkg-spec.yaml, each spec generates an APKG")
parser.add_argument('-t', '--templates-dir', required=True)
parser.add_argument('-o', '--output-dir', required=True, help="directory to put the output apkgs in")
parser.add_argument('-n', '--dry-run', action='store_true', required=False, help="Don't actually build Anki files, but print the files that would be generated")

class NoteType(BaseModel):
    name: str
    fields: list[str]

class CardType(BaseModel):
    name: str
    template: Path

class ImportCsv(BaseModel):
    deck_name_pattern: str
    fields_mapping: list[str]

class ApkgSpec(BaseModel):
    # semver of the content version, appended to the filename after a dash
    content_version: str
    note_type: NoteType
    card_types: list[CardType]
    content: list[dict[Literal['import_apkg'], None]|dict[Literal['import_csv'], ImportCsv]]

    @field_serializer('content_version')
    def dump_content_version(self, ver: Version):
        return str(ver)

    @field_validator('content_version')
    def validate_content_version(cls, version):
        Version.parse(version)
        return version


def load_apkg_spec(spec_path: Path) -> ApkgSpec:
    with open(spec_path) as spec_yaml:
        spec_yaml = load(spec_yaml, Loader=CLoader)
        return ApkgSpec(**spec_yaml)

TEMP_DECK_NAME = 'just_imported'
class ApkgExporter:
    spec: ApkgSpec
    content_dir: Path
    templates_dir: Path
    intermediate_dir: Path
    output_dir: Path
    dry_run: bool
    package_name: str

    def __init__(self, spec: ApkgSpec, content_dir: Path, intermediate_dir: Path, args: argparse.Namespace) -> None:
        self.spec = spec
        self.content_dir = content_dir
        self.templates_dir = Path(args.templates_dir)
        self.intermediate_dir = intermediate_dir
        self.output_dir = Path(args.output_dir)
        self.dry_run = args.dry_run
        self.package_name = content_dir.name


    def export(self) -> None:
        apkg_filename = self.output_dir / f'{self.package_name}-{self.spec.content_version}.apkg'
        if self.dry_run:
            print(str(apkg_filename))
            return

        collection = self.empty_collection()
        note_type = self.add_note_type(collection)
        self.add_content(collection, note_type)
        collection.export_anki_package(
            out_path=str(apkg_filename),
            options=ExportAnkiPackageOptions(
                with_deck_configs=True,
                with_media=True,
                with_scheduling=False,
                legacy=False,
            ),
            limit=None)

    def empty_collection(self) -> Collection:
        collection_filename = self.intermediate_dir / f'{self.package_name}.anki2'
        col = Collection(collection_filename)
        # note: there is only one deck in there with ID 1,
        #       and it cannot be deleted
        self.clear_note_types(col, [])
        return col

    def clear_note_types(self, col: Collection, but_keep_ids: list[int]):
        for model in col.models.all_names_and_ids():
            # one model for the default deck cannot be deleted like this either
            if model.id not in but_keep_ids:
                col.models.remove(model.id)

    def add_note_type(self, col: Collection) -> int:
        note_type_id = 0
        card_types = self.spec.card_types
        name = self.spec.note_type.name
        fields = self.spec.note_type.fields
        return col.models.add_dict({
            'id': note_type_id,
            'name': name,
            'type': 0,
            'mod': 0,
            'usn': 0,
            'sortf': 0,
            'did': None,
            'tmpls': [ self.create_card_type(card_type) for card_type in card_types ],
            'flds': [ self.create_field(idx, field) for idx, field in enumerate(fields) ],
            'css': '',
            'latexPre':'\\documentclass[12pt]{article}\n\\special{papersize=3in,5in}\n\\usepackage[utf8]{inputenc}\n\\usepackage{amssymb,amsmath}\n\\pagestyle{empty}\n\\setlength{\\parindent}{0in}\n\\begin{document}\n',
            'latexPost':'\\end{document}',
            'latexSvg': False,
            'req': [
                [
                    0,
                    'any',
                    [ 0 ]
                ]
            ],
            'originalStockKind': 1
        }).id

    def create_card_type(self, card_type_spec: CardType) -> dict:
        name = card_type_spec.name
        template_dir = self.templates_dir / card_type_spec.template
        front_html_path = template_dir / 'front.html'
        back_html_path = template_dir / 'back.html'
        with open(front_html_path) as front_html_file:
            with open(back_html_path) as back_html_file:
                front_html = front_html_file.read()
                back_html = back_html_file.read()
                return {
                    'name': name,
                    'ord': 0,
                    'qfmt': front_html,
                    'afmt': back_html,
                    'bqfmt': '',
                    'bafmt': '',
                    # this could maybe be the deck override option
                    'did': None,
                    'bfont': '',
                    'bsize': 0,
                    # id?
                }

    def create_field(self, ord: int, field: str) -> dict:
        field_id = random.randint(0, 2 ** 53) # number in safe JSON range
        return {
            'name': field,
            'ord': ord,
            'sticky': False,
            'rtl': False,
            # not sure what the font stuff does, should be styled in CSS
            'font': 'Arial',
            'size': 20,
            'description': '',
            'plainText': False,
            'collapsed': False,
            'excludeFromSearch': False,
            'id': field_id,
            'tag': None,
            'preventDeletion': False
        }

    def add_content(self, col: Collection, note_type_id: int):
        tmp_deck = self.add_tmp_deck(col)
        for content in self.spec.content:
            if 'import_csv' in content:
                import_csv : ImportCsv = content['import_csv']
                deck_name = import_csv.deck_name_pattern
                fields_mapping = import_csv.fields_mapping
                for csv_path in self.content_dir.glob('**/*.csv'):
                    self.add_csv(col, note_type_id, tmp_deck, csv_path, fields_mapping)
                    self.move_to_target_decks(col, tmp_deck, deck_name, csv_path)
            elif 'import_apkg' in content:
                for apkg_path in self.content_dir.glob('**/*.apkg'):
                    self.add_apkg(col, apkg_path, note_type_id)
            else:
                raise Exception(f'I don\'t know how to import this: {content}')
        col.decks.remove([tmp_deck])

    def add_tmp_deck(self, col: Collection) -> int:
        return col.decks.add_normal_deck_with_name(TEMP_DECK_NAME).id

    def get_or_create_deck_id(self, col: Collection, deck_name: str):
        deck = col.decks.by_name(deck_name)
        if deck is None:
            deck_id = col.decks.add_normal_deck_with_name(deck_name).id
            deck = col.decks.get(deck_id)
            if deck is None:
                raise Exception(f'Failed to crate deck {deck_name}')
        return deck['id']

    def add_csv(
            self,
            col: Collection,
            note_type_id: int,
            tmp_deck: int,
            csv_path: Path,
            fields_mapping: list[str]):
        
        mapping = CsvMetadata.MappedNotetype(
            id=note_type_id,
            field_columns=self.field_columns(col, note_type_id, fields_mapping)
        )
        csv_metadata = CsvMetadata(
            delimiter=Delimiter.SEMICOLON,
            is_html=False,
            global_tags=None,
            updated_tags=None,
            # use the fields mapping here?
            column_labels=['' for _ in fields_mapping],
            deck_id=tmp_deck,
            #deck_column=0,
            # for some reason we have to delay the setting, otherwise it has no effect
            global_notetype=mapping,
            #notetype_column=0,#suggested_metadata.notetype_column, # whats this
            tags_column=0,#suggested_metadata.tags_column, # what's this
            force_delimiter=False,
            force_is_html=False,
            preview=[StringList(vals = ['', '', '']), StringList(vals = ['', '', '']), StringList(vals = ['', '', ''])],
            guid_column=0,#suggested_metadata.guid_column, # huh
            dupe_resolution=DupeResolution.UPDATE,
            match_scope=CsvMetadata.MatchScope.NOTETYPE_AND_DECK
        )
        col.import_csv(ImportCsvRequest(
            path=str(csv_path),
            metadata=csv_metadata,
        ))

    def field_columns(self, col: Collection, note_type_id: int, mapping: list[str]) -> list[int]:
        """
        Maps the ordinal of a field against the 1-based index in the CSV.
        Fields that are not found in the CSV are set to 0.
        """
        model_fields = col.models.get(note_type_id)['flds']
        fields : list[int] = [-1] * len(model_fields)
        for model_field in model_fields:
            model_name = model_field['name']
            model_ord = model_field['ord']
            csv_index = mapping.index(model_name) + 1 if model_name in mapping else 0
            fields[model_ord] = csv_index
        return fields

    def add_apkg(self, col: Collection, apkg_path: Path, built_note_type_id: int):
        report = col.import_anki_package(ImportAnkiPackageRequest(
            package_path=str(apkg_path),
            options=ImportAnkiPackageOptions(
                merge_notetypes=True,
                # fail later from log when this happens
                update_notes=ImportAnkiPackageUpdateCondition.IMPORT_ANKI_PACKAGE_UPDATE_CONDITION_ALWAYS,
                # never import note types, we make them ourselves during build
                # anki will map them somehow (from the name?)
                update_notetypes=ImportAnkiPackageUpdateCondition.IMPORT_ANKI_PACKAGE_UPDATE_CONDITION_ALWAYS,
                with_scheduling=False,
                with_deck_configs=False,
            )
        ))

        if report.log.updated:
            raise Exception('unexpected note update')
        if report.log.conflicting:
            raise Exception('unexpected note conflict')
        if report.log.missing_notetype:
            raise Exception('unexpected note missing notetype')
        if report.log.missing_deck:
            raise Exception('unexpected missing deck')
        if report.log.empty_first_field:
            raise Exception('unexpected empty first field')
        if report.log.new:
            new_nids = [ new.id.nid for new in report.log.new ]
            first_note = col.get_note(new_nids[0])
            import_note_type_id = first_note.mid
            info = col.models.change_notetype_info(
                old_notetype_id=import_note_type_id,
                new_notetype_id=built_note_type_id)
            req = ChangeNotetypeRequest(
                note_ids=new_nids,
                new_fields=info.input.new_fields,
                new_templates=info.input.new_templates,
                old_notetype_id=info.input.old_notetype_id,
                new_notetype_id=info.input.new_notetype_id,
                current_schema=info.input.current_schema,
                old_notetype_name=info.input.old_notetype_name,
                is_cloze=info.input.is_cloze
            )
            #req.ParseFromString(info.SerializeToString())
            #req.note_ids.extend(new_nids)
            changes = col.models.change_notetype_of_notes(req)
        # delete models created during import
        self.clear_note_types(col, [built_note_type_id])

    def move_to_target_decks(self,
            col: Collection,
            tmp_deck_id: int,
            deck_name_pattern: str,
            csv_path: Path):
        for card_id in col.find_cards('deck:"' + TEMP_DECK_NAME + '"'):
            card = col.get_card(card_id)
            card_type_name = card.template()["name"]
            desired_deck_name = self.format_deck_name(deck_name_pattern, csv_path, card_type_name)
            card.did = self.get_or_create_deck_id(col, desired_deck_name)
            col.update_card(card)

    def format_deck_name(self, pattern: str, csv_path: Path|None, card_type_name: str|None) -> str:
        replacements = {
            'content_folder': self.package_name
        }
        if csv_path:
            replacements['from_basename'] = csv_path.stem.replace('__', '::')
        if card_type_name:
            replacements['card_type'] = card_type_name
        for key, value in replacements.items():
            pattern = pattern.replace('{{' + key + '}}', value)
        return pattern

def export_package_from_spec(spec_path: Path, args: argparse.Namespace):
    with TemporaryDirectory() as intermediate_dir:
        package_directory = spec_path.parent
        spec = load_apkg_spec(spec_path)
        exporter = ApkgExporter(
            spec=spec,
            content_dir=package_directory,
            intermediate_dir=Path(intermediate_dir),
            args=args
        )
        exporter.export()

args = parser.parse_args()
content = Path(args.content)
if content.suffix == '.yaml' or content.suffix == '.yml':
    # directly specified a YAML as argument
    export_package_from_spec(content, args)
else:
    # got a directory or directory tree to find specs in
    for apkg_spec_path in Path(args.content).glob(f'**/{SPEC_FILENAME}'):
        export_package_from_spec(apkg_spec_path, args)
